// knights on wheels (c) 2008 thomas oppl

#include <default.c>
#include <acknex.h>
#include <stdio.h>
#define KP(key) key_pressed(key)
#define node1_d skill40
#define node1_a skill41
#define node1_f skill42
#define node2_d skill50
#define node2_a skill51
#define node2_f skill52
#define id skill60
#define elapsed_time skill61
#define lap skill62
#define node skill63
#define last_node skill64
#define distance skill65
#define message_time skill66
#define winner skill67
#define energy skill68
#define speed skill69
#define t1_x skill70
#define t1_y skill71
#define t1_z skill72
#define t2_x skill73
#define t2_y skill74
#define t2_z skill75
#define motor_x skill76
#define motor_y skill77
#define motor_z skill78
#define last_position_x skill79
#define last_position_y skill80
#define last_position_z skill81
#define p1_x skill85
#define p1_y skill86
#define p1_z skill87
#define p2_x skill88
#define p2_y skill89
#define p2_z skill90
#define n1_x skill95
#define n1_y skill96
#define n1_z skill97
#define n2_x skill98 
#define n2_y skill99
#define n2_z skill100
#define camdist 128
#define camtilt 28
#define suspension_min 6
#define suspension_max 12
#define beam_length 480
#define beam_force 50000
#define v_mass 16
#define spring 22750
#define damper 425
#define m_speed 100
#define m_power 100
#define m_brake 180
#define steering_power 100
#define max_laps 3
#define e_start 100
#define e_lap 50
#define sd1_distance 320
#define sd1_threshold 640
#define sd2_distance 200
#define sd2_threshold 416
#define v_mode 8

char hud_string[1024];
STRING *str = "#1024";
VIEW *second_camera = {flags = VISIBLE;}
ENTITY *vehicles[2], *wheels[8], *entities[5];
var wheel_ids[8], wheel_i, vehicle_i, i, sd_distance, sd_threshold;
var keys[10] = {72, 77, 80, 75, 17, 32, 31, 30, 83, 16}; 
TEXT *suspensions = {string = "frontleft", "frontright", "rearleft", "rearright";}

void set_physics(h, mh, g, m, f, b, ms, ld, ad)
{
	phent_settype(my, PH_RIGID, h);
	phent_setmass(my, m, mh);
	phent_setfriction(my, f);
	phent_setelasticity(my, b, ms);
	phent_setdamping(my, ld, ad);
	phent_setgroup(my, g);
}

VECTOR *vec_rotate_inverse(VECTOR *v, ANGLE *a)
{
	vec_rotate(v, vector(-a.pan, 0, 0));
	vec_rotate(v, vector(0, -a.tilt, 0));
	vec_rotate(v, vector(0, 0, -a.roll));
	return v;
}

void reset_vehicle(ENTITY *e)
{
	entities[4] = e; 
	
	for(i = 0; i < 5; i++)
	{
		if(i < 4) entities[i] = wheels[i + 4 * e.id];
		phent_enable(entities[i], 0);
		
		vec_add(vec_rotate(vec_rotate_inverse(vec_sub(entities[i].x, e.x), e.pan), vector(e.node2_a, 0, 0)), vector(e.n2_x, e.n2_y, e.n2_z + 64));
		
		vec_set(entities[i].pan, vector(e.node2_a
			- 180 * (i < 4 && (entities[i].id == 1 || entities[i].id == 3 || entities[i].id == 5 || entities[i].id == 7)) 
			, 0, 0));
		
		phent_clearvelocity(entities[i]);
		phent_enable(entities[i], 1);
	}
}

void reset_vehicle_manually()
{
	if(!(key_e && key_pgdn))
	{
		reset_vehicle(
			(long)vehicles[0] * (key_e == 0) +
			(long)vehicles[1] * (key_pgdn == 0)); 	
	}
}

void main()
{	
	add_resource("resource.wrs");
	video_mode = v_mode;
	time_smooth = 0.9;
	video_screen = fog_color = 1;
	camera.fog_end = second_camera.fog_end = 75000;
	
	sd_distance = sd1_distance * (key_f2 == 0) + sd2_distance * (key_f2 == 1);
	sd_threshold = sd1_threshold * (key_f2 == 0) + sd2_threshold * (key_f2 == 1);
	level_load((long)"canyon.wmb" * (key_f2 == 0) + (long)"city.wmb" * (key_f2 == 1));
	
	you = ent_next(0);
	while(you)
	{
		ent_preload(you);
		you = ent_next(you);
	}
	
	camera.x = camera.y = camera.tilt = second_camera.x = second_camera.y = second_camera.tilt = wheel_i = vehicle_i = 0; 
	camera.pan = camera.z = second_camera.pan = second_camera.z = 90;
	sky_cube_level.material = mat_unlit;
	on_e = on_pgdn = reset_vehicle_manually;
	on_f1 = on_f2 = main;
	ph_setgravity(vector(0, 0, -320 * 1.4));
	ph_setcorrections(25000, 0.05);
}

action vehicle()
{
	my.id = vehicle_i++ - 1;
	vehicles[my.id] = my;
	my.energy = e_start;

	set_physics(
		PH_BOX,
		PH_BOX,
		2 + (my.id == 1) * 2, 
		v_mass,
		40,
		25,
		10,
		20,
		20
		);
	
	VIEW *my_camera = (long)second_camera * (my.id == 0) + (long)camera * (my.id == 1);
	
	path_set(my, "driveline");
	
	while(1)
	{	
		// shortcut detection
		if(my.lap < max_laps) 
		{	
			my.elapsed_time += time_step / 16;		
			
			if((my.node = path_scan(my, my.x, my.pan, vector(360, 0, sd_distance))) != 0) 
			{
				path_getnode(my, my.node, my.n1_x, my.node1_d);
				path_getnode(my, my.node * (my.last_node == 0) + my.last_node, my.n2_x, my.node2_d);
				
				if(my.node1_f == 1 && my.node2_f == 2)
				{
					my.lap++;
					my.energy = minv(100, my.energy + e_lap);
					if(my.lap == max_laps)
					{
						if(vehicles[1 - my.id].winner == 0) my.winner = 1;
					}
				}
				
				if((my.node1_f == 2 && my.node2_f == 1) 
				|| (my.distance != 0 && my.distance < (my.node1_d - my.node2_d - sd_threshold))) 
				{
					my.message_time = my.elapsed_time + 3; 
					reset_vehicle(my);
				}
				else
				{			
					my.last_node = my.node; 
				}
				my.distance = 0;
			}
			else
			{
				if(vec_dist(my.x, my.last_position_x) > 0.1) my.distance += vec_dist(my.x, my.last_position_x); 
			}
			vec_set(my.last_position_x, my.x);
		}
		
		// hud
		camera.size_x = second_camera.size_x = second_camera.pos_x = screen_size.x / 2;
		sprintf(hud_string,
			"player %d\ntime: %02d:%02d:%02d\nspeed: %dkm/h\nlap: %d\nenergy: %d%%\n\n%s",
			(int)my.id + 1,
			(int)my.elapsed_time / 60, (int)my.elapsed_time % 60, (int)(fraction(my.elapsed_time) * 100),
			(int)(my.speed * 1.2),
			(int)clamp(my.lap + 1, 1, max_laps),
			(int)my.energy,
			(long)"" * (my.message_time <= my.elapsed_time && my.winner == 0) +
			(long)"invalid shortcut!" * (my.message_time > my.elapsed_time && my.winner == 0) +
			(long)"winner!" * (my.winner == 1)
			);
		draw_text(hud_string, screen_size.x / 2 + 20 - screen_size.x / 2 * my.id, 20, vector(0,0,0));
		
		// controls
		phcon_setmotor(wheel_ids[2 + 4 * my.id], nullvector,
			vec_set(my.motor_x, vector(
				m_speed * KP(keys[0 + 4 * my.id]) * !KP(keys[2 + 4 * my.id]),
				m_power * KP(keys[0 + 4 * my.id]) * !KP(keys[2 + 4 * my.id]) + m_brake * KP(keys[2 + 4 * my.id]),
				0)), nullvector);
		phcon_setmotor(wheel_ids[3 + 4 * my.id], nullvector, my.motor_x, nullvector);
		
		phcon_getposition(wheel_ids[0 + 4 * my.id], my.p1_x);
		phcon_setmotor(wheel_ids[0 + 4 * my.id],
			vector(((30 * (KP(keys[1 + 4 * my.id]) - KP(keys[3 + 4 * my.id]))) - my.p1_x) * minv(time_step, 1), steering_power, 0),
			my.motor_x, nullvector);
		
		phcon_getposition(wheel_ids[1 + 4 * my.id], my.p2_x);
		phcon_setmotor(wheel_ids[1 + 4 * my.id],
			vector(((30 * (KP(keys[1 + 4 * my.id]) - KP(keys[3 + 4 * my.id]))) - my.p2_x) * minv(time_step, 1), steering_power, 0),
			my.motor_x, nullvector);
		
		// suspension
		ent_animate(my, "steering", clamp((my.p1_x + my.p2_x) / 2 * (100.0 / 60.0) + 50, 0, 100), 0);
		
		// camera
		if(!(def_camera && my.id == 1))
		{
			vec_add(my_camera.x, vector(
				((my.x + cosv(my.pan) * fcos(-camtilt, -camdist)) - my_camera.x) * minv(0.4 * time_step, 1),
				((my.y + sinv(my.pan) * fcos(-camtilt, -camdist)) - my_camera.y) * minv(0.4 * time_step, 1),
				((my.z + fsin(-camtilt, -camdist)) - my_camera.z) * minv(0.4 * time_step, 1)
				));
			
			if(absv(my.pan - my_camera.pan) > 180) my_camera.pan += 360 * sign(my.pan - my_camera.pan);
			
			my_camera.pan += (my.pan - my_camera.pan) * minv(0.4 * time_step, 1);
			my_camera.tilt = -camtilt * 0.5;
		}
		
		// tractor beam
		if(((my == vehicles[0] && KP(keys[8]))
		|| (my == vehicles[1] && KP(keys[9])))
		&& my.energy > 0)
		{
			my.energy = maxv(0, my.energy - time_step);
			
			c_trace(vec_set(my.p1_x, vec_for_vertex(my.t1_x, my, 1646)),
				vec_set(my.p2_x, vec_add(vec_scale(vec_for_angle(my.t2_x, my.pan), beam_length), my.p1_x)),
				IGNORE_ME | IGNORE_FLAG2);
			
			if(trace_hit) vec_set(my.p2_x, target);
			if(you == vehicles[1 - my.id]) phent_addforceglobal(you, vec_inverse(vec_scale(vec_for_angle(my.t2_x, my.pan), beam_force)), target);
			
			for(i = 0; i < integer(vec_dist(my.p1_x, my.p2_x)); i++)
			{
				vec_add(my.t1_x, vec_add(vec_for_angle(my.t2_x, my.pan), vector(random(0.8)-0.4, random(0.8)-0.4, random(0.8)-0.4)));
				draw_point3d(my.t1_x , vector(255 * (my.id == 0), 0, 255 * (my.id == 1)), 100, 1.5);
			}
		}
		
		wait(1);
	}
}

action tire()
{
	my.id = wheel_i++ - 1;
	wheels[my.id] = my;
	
	set_physics(
		PH_SPHERE,
		PH_SPHERE,
		2 + (my.id > 3) * 2, 
		30,
		80,
		50,
		10,
		20,
		20
		);

	wheel_ids[my.id] = phcon_add(PH_WHEEL, vehicles[my.id > 3], my); 
	phcon_setparams1(wheel_ids[my.id], my.x, vector(0, 0, 1), vector(1, 0, 0));
	phcon_setparams2(wheel_ids[my.id], vector(
		-30 * (my.id == 0 || my.id == 1 || my.id == 4 || my.id == 5), 
		30 * (my.id == 0 || my.id == 1 || my.id == 4 || my.id == 5), 
		0), nullvector, vector(spring, damper, 0));
	
	while(1)
	{
		// suspension
		ent_animate(vehicles[my.id > 3], (suspensions.pstring)[my.id % 4],
			clamp((100.0 / (suspension_max - suspension_min)) * (
			(vec_rotate_inverse(vec_diff(my.t1_x, my.x, vehicles[my.id > 3].x), vehicles[my.id > 3].pan)).z 
			- suspension_min) - 30, 0, 100), ANM_ADD);
		
		phent_getvelocity(my, my.t2_x, nullvector);
		vehicles[my.id > 3].speed = (vec_length(my.t2_x) / 32.0) * 60 * 60 / 1000.0;
		
		wait(1);
	}
}

action water()
{
	while(1)
	{
		my.v += 0.8 * time_step;
		wait(1);
	}	
}